#Arquivo de Leitura das Entradas do Problema
from reader import *
#Modulo Operator do Python
import operator

######################--VALIDACAO DE RESTRIcOES--############################
def is_a_valid_solution(solution):

    return is_a_valid_solution_out(solution, intolerable_warning_point, project_duration, warning_points_hash, costs, resources, mandatory_project, generator_unit, halting_month, halting_duration, project_powerplant, T)

#end def

def is_a_valid_solution_out(solution, intolerable_warning_point, project_duration, warning_points_hash, costs, resources, mandatory_project, generator_unit, halting_month, halting_duration, project_powerplant, T):
 
 return (check_consistent_scheduling(solution, T) 
        and check_mandatory_project(solution, mandatory_project) 
        and check_limited_resource(solution, costs, resources, 'CAPEX', T)
        and check_limited_resource(solution, costs, resources, 'OPEX', T)
        and check_halting_location_RP(solution, maintenance, project_duration, RP, generator_unit, halting_month, halting_duration, project_powerplant, T)
        and check_halting_location_BBB(solution, maintenance, project_duration, BBB, T)
        and check_halting_location_BAR(solution, maintenance, project_duration, T) 
        and check_halting_location_BAB(solution, maintenance, project_duration, T)
        and check_halting_location_NAPI(solution, maintenance, project_duration, NAPI, T))

#end def 
######################--VALIDACAO DE RESTRIcOES--#####################-END-##

#######################--RESTRIÇÃO DE AGENDAMENTO--##########################
def check_consistent_scheduling(solution, T):

    for k in solution.keys():
        if type(solution[k]) is not int:
            return False
        #end if
        
        if solution[k] > 2*T or solution[k] < 1:
            return False
        #end if
    #end for
        
    return True
#end def
#######################--RESTRIÇÃO DE AGENDAMENTO--###################-END-##

################--RESTRIÇÃO DE PROJETOS OBRIGAToRIOS--#######################
def check_mandatory_project(solution, mandatory_project):

    #Para cada projeto k
    for k in mandatory_project.keys():
        #Compara se o mes de k esta igual ao mes da solucAo para k.
        if mandatory_project[k] != solution[k]:
            return False
        #end if
    #end for

    return True        
#end def
################--RESTRIÇÃO DE PROJETOS OBRIGAToRIOS--################-END-##

#####################--FUNÇÃO AUXILIAR DE CUSTO--############################
def aux_costs(project, initial_month, current_month, costs, project_duration):

    if current_month >= initial_month and current_month <= initial_month + project_duration[project] - 1:

        return costs[project][current_month-initial_month]

    else:
        return 0
    #end if
#enf def
#####################--FUNÇÃO AUXILIAR DE CUSTO--#####################-END-##

##################--RESTRIÇÃO DE RECURSOS LIMITADOS--########################
def check_limited_resource(solution, costs, resources, resource_category, T):

    for year in range(int(T/12)):

        total_year_costs = 0.0

        for month in range(1,13):

            current_month =  12*year + month

            for project in solution.keys():

                if (resource_category == category[project]):

                    total_year_costs += aux_costs(project, solution[project], current_month, costs, project_duration)

                #end if

            #end for

        #end for

        if total_year_costs > resources[resource_category][year]:

            return False

        #end if

    #end for

    return True

#enf def
###################--RESTRIÇÃO DE RECURSOS LIMITADOS--################-END-##

##############--RESTRIÇÃO DE PONTOS DE ATENÇÃO INTOLERAVEIS--################
def check_intolerable_warning_point(solution):

    return check_intolerable_warning_point_out(solution, intolerable_warning_point, project_duration, warning_points_hash)

#end def

##############--RESTRIÇÃO DE PONTOS DE ATENÇÃO INTOLERAVEIS--#########-END-################--RESTRIÇÃO DE PONTOS DE ATENÇÃO INTOLERAVEIS--################
def check_intolerable_warning_point_out(solution, intolerable_warning_point, project_duration, warning_points_hash):

    for wp in intolerable_warning_point.keys():

        meses_finais = [ (solution[project] + project_duration[project] -1) for project in warning_points_hash[wp]]

        last_month = max(meses_finais)

        if (last_month > intolerable_warning_point[wp]):

            return False

        # end if

    # end for

    return True

#end def
##############--RESTRIÇÃO DE PONTOS DE ATENÇÃO INTOLERAVEIS--#########-END-##

###################--RESTRIÇÃO PARADA POR LOCALIZAcAO RP--###################
def check_halting_location_RP(solution, maintenance, project_duration, RP, generator_unit, halting_month, halting_duration, project_powerplant, T):

    #Dicionário interno de usinas da FUNÇÃO check_halting_location_RP
    g = {}
    
    #Variavel que contém o maior valor de unidades geradoras paradas
    max_halting_unit_generation = 0

    #Para cada mês dentro do horizonte de planejamento
    for month in range(2*T):

        #Para cada usina na Localidade de RP
        for powerplant in RP:

            #Recebe valor 0 de unidades geradoras de energia paradas
            g[powerplant] = 0

        #end for

        #Para cada projeto na Solução
        for project in solution.keys():

            #Se o projeto estiver contido o dicionário Maintenance
            if project in maintenance.keys():

                if (solution[project] <= month and month <= solution[project] + project_duration[project] -1):
                    
                    #Variavel com o Mês inicial da parada
                    initial_halting_month = solution[project] + int(halting_month[project])

                    #Variavel com o Mês final da parada
                    last_halting_month = initial_halting_month + int(halting_month[project]) -1

                    #Se
                    if (initial_halting_month <= month and last_halting_month >= month):

                        #Para cada usina em RP
                        for powerplant in RP:

                            if project_powerplant[project] == powerplant:

                                g[powerplant] += 1

                            #enf if

                        #end for

                    #enf if

                #enf if

             #end if

        #end for

        max_halting_unit_generation = max(g.items(), key= lambda g: g[1])

        #Compara o se o maior valor de unidades geradoras paradas é maior ou igual a 2
        if (max_halting_unit_generation[1] >= 2):

            for pp in RP - max_halting_unit_generation:

                if (g[powerplant] > 0):

                    return False

                #end if

            #end for

        #end if

    return True

    #end for

#end def
###################--RESTRIÇÃO PARADA POR LOCALIZAcAO RP--############-END-##

#################--RESTRIÇÃO PARADA POR LOCALIZAcAO BBB--####################
def check_halting_location_BBB(solution, maintenance, project_duration, BBB, T):

    g = {}

    for month in range(2*T):

        for powerplant in BBB:

            g[powerplant] = 0

        #end for

        for project in solution.keys():

            if project in maintenance.keys():

                if (solution[project] <= month and month <= solution[project] + project_duration[project] -1):

                    initial_halting_month = solution[project] + int(halting_month[project])

                    last_halting_month = initial_halting_month + int(halting_month[project]) -1

                    if (initial_halting_month <= month and last_halting_month >= month):

                        for powerplant in BBB:

                            if (project_powerplant[project] == powerplant):

                                g[powerplant] += 1

                            #enf if

                        #end for

                    #enf if
 
                #enf if

            #end if

        #end if

        twoplusHaltingGUs = [ pp for pp in g.keys() if g[pp] >= 2 ]

        if (twoplusHaltingGUs != []):

            for pp in list(set(BBB) - set(twoplusHaltingGUs[0])):

                if (g[pp] > 0):

                    return False

                #end if

            #end for

        #end if

    return True

    #end for

#enf def
#################--RESTRIÇÃO PARADA POR LOCALIZAcAO BBB--#############-END-##

#################--RESTRIÇÃO PARADA POR LOCALIZAcAO BAR--####################
def check_halting_location_BAR(solution, maintenance, project_duration, T):

    g = {}
    
    for month in range(2*T):

        for powerplant in BBB:

            g[powerplant] = 0

        #end for

        for project in solution.keys():

            if project in maintenance.keys():

                if (solution[project] <= month and month <= solution[project] + project_duration[project] -1):

                    initial_halting_month = solution[project] + int(halting_month[project])

                    last_halting_month = initial_halting_month + int(halting_month[project]) -1

                    if (initial_halting_month <= month and last_halting_month >= month):

                        for powerplant in BBB:

                            if project_powerplant[project] == powerplant:

                                g[powerplant] += 1

                            #enf if

                        #end for

                    #enf if

                #enf if

             #end if

         #end if

        twoplusHaltingGUs = [ pp for pp in g.keys() if g[pp] >= 2 ]

        if (twoplusHaltingGUs != []):

            for pp in list(set(BAR) - set(twoplusHaltingGUs[0])):

                if (g[pp] > 0):

                    return False

                #end if

            #end for

        #end if

    return True

    #end for

#enf def
#################--RESTRIÇÃO PARADA POR LOCALIZAcAO BAR--#############-END-##
                               
#################--RESTRIÇÃO PARADA POR LOCALIZAcAO BAB--####################
def check_halting_location_BAB(solution, maintenance, project_duration, T):

    g = {}
    
    for month in range(2*T):

        for powerplant in BBB:

            g[powerplant] = 0

        #end for

        for project in solution.keys():
            
            if project in maintenance.keys():
                
                if (solution[project] <= month and month <= solution[project] + project_duration[project] -1):

                    initial_halting_month = solution[project] + int(halting_month[project])

                    last_halting_month = initial_halting_month + int(halting_month[project]) -1

                    if (initial_halting_month <= month and last_halting_month >= month):

                        for powerplant in BBB:

                            if project_powerplant[project] == powerplant:

                                g[powerplant] += 1

                            #enf if

                        #end for

                    #enf if

                #enf if

             #end if

         #end if

        twoplusHaltingGUs = [ pp for pp in g.keys() if g[pp] >= 2 ]

        if (twoplusHaltingGUs != []):

            for pp in list(set(BAB, BAR) - set(twoplusHaltingGUs[0])):

                if (g[pp] > 0 and g[pp] >= 2):

                    return False

                #end if

            #end for

        #end if

        return True

    #end for

#enf def
#################--RESTRIÇÃO PARADA POR LOCALIZAcAO BAB--#############-END-##

################--RESTRIÇÃO PARADA POR LOCALIZAcAO NAPI--####################
def check_halting_location_NAPI(solution, maintenance, project_duration, NAPI, T):
    
    g = {}
    
    for month in range(2*T):

        for powerplant in NAPI:

            g[powerplant] = 0

        #end for

        for project in solution.keys():

            if project in maintenance.keys():

                if (solution[project] <= month and month <= solution[project] + project_duration[project] -1):

                    initial_halting_month = solution[project] + int(halting_month[project])

                    last_halting_month = initial_halting_month + int(halting_month[project]) -1

                    if (initial_halting_month <= month and last_halting_month >= month):

                        for powerplant in NAPI:

                            if project_powerplant[project] == powerplant:

                                g[powerplant] += 1

                            #enf if

                        #end for

                    #enf if

                #enf if

             #end if

         #end if

        halting_unit_generator = 0

        for powerplant in NAPI:

            halting_unit_generator = g[powerplant]

        #end for

        if halting_unit_generator > 2:

            return False

        #end if

    return True

    #end for

#enf def
################--RESTRIÇÃO PARADA POR LOCALIZAcAO NAPI--#############-END-##

#################--RESTRIÇÃO PARADA POR LOCALIZAcAO AGV--####################
def check_halting_location_AGV(solution, maintenance, project_duration, T):

    g = {}
    
    for month in range(2*T):

        for powerplant in AV:

            g[powerplant] = 0

        #end for

        for project in solution.keys():
            
            if project in maintenance.keys():
                
                if (solution[project] <= month and month <= solution[project] + project_duration[project] -1):

                    initial_halting_month = solution[project] + int(halting_month[project])

                    last_halting_month = initial_halting_month + int(halting_month[project]) -1

                    if (initial_halting_month <= month and last_halting_month >= month):

                        for powerplant in AV:

                            if project_powerplant[project] == powerplant:
                                print(project)
                                g[powerplant] += 1

                            #enf if

                        #end for

                    #enf if

                #enf if

             #end if

         #end if

        halting_unit_generator = 0

        for powerplant in AV:

            halting_unit_generator = g[powerplant]

        #end for

        if halting_unit_generator > 2:

            return False

        #end if

    return True

    #end for

#enf def
#################--RESTRIÇÃO PARADA POR LOCALIZAcAO AGV--#############-END-##